package gov.usgs.earthquake.nshmp.netcdf.www;

import java.nio.file.Path;

import gov.usgs.earthquake.nshmp.gmm.NehrpSiteClass;
import gov.usgs.earthquake.nshmp.netcdf.NetcdfGroundMotions;
import gov.usgs.earthquake.nshmp.netcdf.www.NetcdfService.RequestDataSiteClass;
import gov.usgs.earthquake.nshmp.netcdf.www.NetcdfService.ResponseData;
import gov.usgs.earthquake.nshmp.netcdf.www.NetcdfService.ServiceResponseMetadata;
import gov.usgs.earthquake.nshmp.www.NshmpMicronautServlet;
import gov.usgs.earthquake.nshmp.www.ResponseBody;

import io.micronaut.context.annotation.Value;
import io.micronaut.context.event.StartupEvent;
import io.micronaut.core.annotation.Nullable;
import io.micronaut.http.HttpRequest;
import io.micronaut.http.HttpResponse;
import io.micronaut.http.MediaType;
import io.micronaut.http.annotation.Controller;
import io.micronaut.http.annotation.Get;
import io.micronaut.http.annotation.PathVariable;
import io.micronaut.http.annotation.QueryValue;
import io.micronaut.runtime.event.annotation.EventListener;
import io.swagger.v3.oas.annotations.Hidden;
import io.swagger.v3.oas.annotations.Operation;
import io.swagger.v3.oas.annotations.media.Content;
import io.swagger.v3.oas.annotations.media.Schema;
import io.swagger.v3.oas.annotations.responses.ApiResponse;
import io.swagger.v3.oas.annotations.tags.Tag;
import jakarta.inject.Inject;

/**
 * Micronaut controller for getting ground motions from a NetCDF file.
 *
 * @see NetcdfService
 *
 * @author U.S. Geological Survey
 */
@Tag(name = "Web Service")
@Controller("/spectra")
public class NetcdfController {

  @Inject
  private NshmpMicronautServlet servlet;

  @Value("${nshmp-ws-static.netcdf-file}")
  Path netcdfPath;

  NetcdfServiceGroundMotions service;

  /**
   * Read in data type and return the appropriate service to use.
   */
  @EventListener
  void startup(StartupEvent event) {
    var netcdf = new NetcdfGroundMotions(netcdfPath);
    service = new NetcdfServiceGroundMotions(netcdf);
  }

  /**
   * GET method to return static curves using slash delimited.
   *
   * @param request The HTTP request
   * @param longitude Longitude of site, in decimal degrees
   * @param latitude Latitude of site, in decimal degrees
   * @param siteClass Site class
   */
  @Operation(
      summary = "Returns risk-targeted design response spectra from a slash based call",
      description = "### Returns a risk-targeted design response spectrum for a " +
          "user-specified latitude, longitude, and site class.\n" +

          "Enter the latitude and longitude select the site class, and press `Execute`.\n" +

          "### Service call pattern\n" +
          "This service call is slashed delimited with pattern: " +
          "`/spectra/{longitude}/{latitude}/{siteClass}`\n" +
          "<br><br>" +
          "Example: `/spectra/-118/34/A`",
      operationId = "aashto-slash")
  @ApiResponse(
      description = "Spatially interpolates data from https://doi.org/10.5066/P9Z206HY",
      responseCode = "200",
      content = @Content(
          mediaType = MediaType.APPLICATION_JSON,
          schema = @Schema(implementation = Response.class)))
  @Get(uri = "/{longitude}/{latitude}/{siteClass}", produces = MediaType.APPLICATION_JSON)
  public HttpResponse<String> doGetSlashBySite(
      HttpRequest<?> request,
      @Schema(required = true) @PathVariable @Nullable Double latitude,
      @Schema(required = true) @PathVariable @Nullable Double longitude,
      @Schema(required = true) @PathVariable @Nullable NehrpSiteClass siteClass) {
    var query = new Query(longitude, latitude, siteClass);
    return service.handleServiceCall(request, query);
  }

  /**
   * GET method to return a static curve using URL query.
   *
   * @param request The HTTP request
   * @param longitude The longitude of the site
   * @param latitude Latitude of the site
   * @param siteClass The site class (optional)
   */
  @Hidden
  @Operation(
      summary = "Returns risk-targeted design response spectra from a query based call",
      description = "### Returns a risk-targeted design response spectrum for a " +
          "user-specified latitude, longitude, and site class.\n" +

          "Enter the latitude and longitude select the site class, and press `Execute`.\n" +

          "### Service call pattern\n" +
          "This service call is query based with pattern: " +
          "`/spectra?longitude={number}&latitude={number}&siteClass={string}`\n" +
          "<br><br>" +
          "Example: `/spectra?longitude=-118&latitude=34&siteClass=A`",
      operationId = "aashto-query")
  @ApiResponse(
      description = "Spatially interpolates data from https://doi.org/10.5066/P9Z206HY",
      responseCode = "200",
      content = @Content(
          mediaType = MediaType.APPLICATION_JSON,
          schema = @Schema(implementation = Response.class)))
  @Get(uri = "{?longitude,latitude,siteClass}", produces = MediaType.APPLICATION_JSON)
  public HttpResponse<String> doGet(
      HttpRequest<?> request,
      @Schema(required = true) @QueryValue @Nullable Double latitude,
      @Schema(required = true) @QueryValue @Nullable Double longitude,
      @QueryValue @Nullable NehrpSiteClass siteClass) {
    var query = new Query(longitude, latitude, siteClass);
    return service.handleServiceCall(request, query);
  }

  /**
   * GET method to return hazard curves using slash delimited.
   *
   * @param request The HTTP request
   * @param longitude The longitude of the site
   * @param latitude Latitude of the site
   */
  @Hidden
  @Get(uri = "/{longitude}/{latitude}", produces = MediaType.APPLICATION_JSON)
  public HttpResponse<String> doGetSlash(
      HttpRequest<?> request,
      @Schema(required = true) @PathVariable @Nullable Double longitude,
      @Schema(required = true) @PathVariable @Nullable Double latitude) {
    var query = new Query(longitude, latitude, null);
    return service.handleServiceCall(request, query);
  }

  // For Swagger schema
  private static class Response
      extends
      ResponseBody<RequestDataSiteClass, ResponseData<ServiceResponseMetadata>> {}
}
