package gov.usgs.earthquake.nshmp.netcdf.www;

import java.util.List;
import java.util.stream.Collectors;

import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.gmm.Imt;
import gov.usgs.earthquake.nshmp.gmm.NehrpSiteClass;
import gov.usgs.earthquake.nshmp.netcdf.NetcdfHazardCurves;
import gov.usgs.earthquake.nshmp.netcdf.NetcdfVersion;
import gov.usgs.earthquake.nshmp.netcdf.data.StaticData;
import gov.usgs.earthquake.nshmp.netcdf.data.StaticDataHazardCurves;
import gov.usgs.earthquake.nshmp.netcdf.www.NetcdfWsUtils.Key;
import gov.usgs.earthquake.nshmp.netcdf.www.Query.Service;
import gov.usgs.earthquake.nshmp.www.ResponseBody;
import gov.usgs.earthquake.nshmp.www.ResponseMetadata;
import gov.usgs.earthquake.nshmp.www.WsUtils;

import io.micronaut.http.HttpRequest;

/**
 * Handles service calls for hazard curves.
 *
 * @see NetcdfController
 *
 * @author U.S. Geological Survey
 */
public class NetcdfServiceHazardCurves extends NetcdfService<HazardQuery> {

  static final String SERVICE_DESCRIPTION = "Get static hazard curves from a NetCDF file";
  static final String SERVICE_NAME = "Static Hazard Curves";
  static final String X_LABEL = "Ground Motion (g)";
  static final String Y_LABEL = "Annual Frequency of Exceedence";

  public NetcdfServiceHazardCurves(NetcdfHazardCurves netcdf) {
    super(netcdf);
  }

  @Override
  ResponseBody<String, Metadata> getMetadataResponse(HttpRequest<?> request) {
    var metadata = new Metadata(request, SERVICE_DESCRIPTION);
    return ResponseBody.<String, Metadata> usage()
        .metadata(new ResponseMetadata(NetcdfVersion.appVersions()))
        .name(SERVICE_NAME)
        .request(NetcdfWsUtils.getRequestUrl(request))
        .response(metadata)
        .url(NetcdfWsUtils.getRequestUrl(request))
        .build();
  }

  @Override
  public Service getService(HazardQuery query) {
    if (query.imt != null) {
      return Service.CURVES_BY_IMT;
    } else if (query.siteClass != null) {
      return Service.CURVES_BY_SITE_CLASS;
    } else {
      return Service.CURVES;
    }
  }

  @Override
  String getServiceName() {
    return SERVICE_NAME;
  }

  @Override
  SourceModel getSourceModel() {
    return new HazardSourceModel();
  }

  @Override
  NetcdfHazardCurves netcdf() {
    return (NetcdfHazardCurves) netcdf;
  }

  @Override
  ResponseBody<RequestData, List<List<ResponseData<HazardResponseMetadata>>>> processCurves(
      RequestData request,
      String url) {
    WsUtils.checkValue(Key.LATITUDE, request.latitude);
    WsUtils.checkValue(Key.LONGITUDE, request.longitude);
    var curves = netcdf().staticData(request.site);
    var curvesAsList = toList(request.site, curves);

    return ResponseBody.<RequestData, List<List<ResponseData<HazardResponseMetadata>>>> success()
        .metadata(new ResponseMetadata(NetcdfVersion.appVersions()))
        .name(SERVICE_NAME)
        .request(request)
        .response(curvesAsList)
        .url(url)
        .build();
  }

  @Override
  ResponseBody<RequestDataSiteClass, List<ResponseData<HazardResponseMetadata>>> processCurvesSiteClass(
      RequestDataSiteClass request,
      String url) {
    WsUtils.checkValue(Key.LATITUDE, request.latitude);
    WsUtils.checkValue(Key.LONGITUDE, request.longitude);
    WsUtils.checkValue(Key.SITE_CLASS, request.siteClass);
    var curves = netcdf().staticData(request.site, request.siteClass);
    var curvesAsList = toList(request, curves);

    return ResponseBody.<RequestDataSiteClass, List<ResponseData<HazardResponseMetadata>>> success()
        .metadata(new ResponseMetadata(NetcdfVersion.appVersions()))
        .name(SERVICE_NAME)
        .request(request)
        .response(curvesAsList)
        .url(url)
        .build();
  }

  ResponseBody<RequestDataSiteClass, ResponseData<HazardResponseMetadata>> processCurvesImt(
      RequestDataImt request,
      String url) {
    WsUtils.checkValue(Key.LATITUDE, request.latitude);
    WsUtils.checkValue(Key.LONGITUDE, request.longitude);
    WsUtils.checkValue(Key.SITE_CLASS, request.siteClass);
    WsUtils.checkValue(Key.IMT, request.imt);
    var curves = netcdf().staticData(request.site, request.siteClass);

    return ResponseBody.<RequestDataSiteClass, ResponseData<HazardResponseMetadata>> success()
        .metadata(new ResponseMetadata(NetcdfVersion.appVersions()))
        .name(SERVICE_NAME)
        .request(request)
        .response(toResponseData(request, request.imt, curves.get(request.imt)))
        .url(url)
        .build();
  }

  @Override
  ResponseBody<?, ?> processRequest(HttpRequest<?> httpRequest, HazardQuery query,
      Service service) {
    var site = Location.create(query.longitude, query.latitude);
    var url = NetcdfWsUtils.getRequestUrl(httpRequest);

    switch (service) {
      case CURVES:
        var requestData = new RequestData(site);
        return processCurves(requestData, url);
      case CURVES_BY_SITE_CLASS:
        var requestDataSiteClass = new RequestDataSiteClass(site, query.siteClass);
        return processCurvesSiteClass(requestDataSiteClass, url);
      case CURVES_BY_IMT:
        var requestDataImt = new RequestDataImt(site, query.siteClass, query.imt);
        return processCurvesImt(requestDataImt, url);
      default:
        throw new RuntimeException("Netcdf service [" + service + "] not found");
    }
  }

  List<ResponseData<HazardResponseMetadata>> toList(
      RequestDataSiteClass request,
      StaticDataHazardCurves curves) {
    return curves.entrySet().stream()
        .map((entry) -> toResponseData(request, entry.getKey(), entry.getValue()))
        .collect(Collectors.toList());
  }

  List<List<ResponseData<HazardResponseMetadata>>> toList(
      Location site,
      StaticData<StaticDataHazardCurves> curves) {
    return curves.entrySet().stream()
        .map(entry -> {
          var request = new RequestDataSiteClass(site, entry.getKey());
          return toList(request, entry.getValue());
        })
        .collect(Collectors.toList());
  }

  ResponseData<HazardResponseMetadata> toResponseData(
      RequestDataSiteClass request,
      Imt imt,
      XySequence curves) {
    var metadata =
        new HazardResponseMetadata(request.siteClass, imt, X_LABEL, Y_LABEL);
    return new ResponseData<>(metadata, curves);
  }

  class HazardSourceModel extends SourceModel {
    public final List<Imt> imts;

    HazardSourceModel() {
      super();
      imts = netcdf().netcdfData().imts().stream().sorted().collect(Collectors.toList());
    }
  }

  static class HazardResponseMetadata extends ServiceResponseMetadata {
    public final Imt imt;

    HazardResponseMetadata(NehrpSiteClass siteClass, Imt imt, String xLabel, String yLabel) {
      super(siteClass, xLabel, yLabel);
      this.imt = imt;
    }
  }
}
