package gov.usgs.earthquake.nshmp.netcdf.data;

import static com.google.common.base.Preconditions.checkState;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

/**
 * Create NetCDF shapes and keep track of indices.
 *
 * @author U.S. Geological Survey
 */
public class NetcdfShape {

  private final Map<IndexKey, Integer> indexMap;

  private NetcdfShape(Builder builder) {
    indexMap = builder.indexMap;
  }

  /**
   * Returns builder to build a NetCDF shape.
   */
  public static Builder builder() {
    return new Builder();
  }

  /**
   * Returns a builder to build an array shape for NetCDF origin and shapes.
   */
  public BuildShape buildShape() {
    return new BuildShape();
  }

  public class BuildShape {
    List<KeySize> keySizes;
    Optional<Integer> pad = Optional.empty();
    boolean reduce = false;

    private BuildShape() {
      keySizes = new ArrayList<>();
    }

    /**
     * Add key and associated size for shape.
     *
     * @param key The key
     * @param size The size
     */
    public BuildShape add(IndexKey key, int size) {
      checkState(indexMap.containsKey(key), String.format("Key [%s] not found in shape", key));
      keySizes.add(new KeySize(key, size));
      return this;
    }

    /**
     * Change to padding value of 0.
     *
     * Not used if reducing shape.
     *
     * @param value The value to pad the array shape
     */
    public BuildShape pad(int value) {
      pad = Optional.of(value);
      return this;
    }

    /**
     * Reduce the array to only added keys in builder.
     */
    public BuildShape reduce() {
      reduce = true;
      return this;
    }

    /**
     * Returns the shape array.
     */
    public int[] build() {
      if (reduce) {
        var reducePad = -1;
        return build(reducePad).stream()
            .mapToInt(Integer::intValue)
            .filter(val -> val != reducePad)
            .toArray();
      } else {
        return build(pad.orElse(0)).stream().mapToInt(Integer::intValue).toArray();
      }
    }

    private List<Integer> build(int pad) {
      var shape = new ArrayList<>(Collections.nCopies(indexMap.size(), pad));
      keySizes.forEach(keySize -> {
        shape.set(indexMap.get(keySize.indexKey), keySize.size);
      });

      return shape;
    }
  }

  public static class KeySize {
    public final IndexKey indexKey;
    public final int size;

    private KeySize(IndexKey indexKey, int size) {
      this.indexKey = indexKey;
      this.size = size;
    }
  }

  public static enum IndexKey {
    IML,
    IMT,
    LATITUDE,
    LONGITUDE,
    SITE_CLASS;
  }

  public static class Builder {
    Map<IndexKey, Integer> indexMap;

    private Builder() {
      indexMap = new HashMap<>();
    }

    public Builder add(IndexKey key, int index) {
      indexMap.put(key, index);
      return this;
    }

    public NetcdfShape build() {
      return new NetcdfShape(this);
    }
  }
}
