package gov.usgs.earthquake.nshmp.netcdf.reader;

import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;

import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.LocationList;
import gov.usgs.earthquake.nshmp.netcdf.Netcdf;
import gov.usgs.earthquake.nshmp.netcdf.data.BoundingData;
import gov.usgs.earthquake.nshmp.netcdf.data.NetcdfData;
import gov.usgs.earthquake.nshmp.netcdf.data.NetcdfShape;
import gov.usgs.earthquake.nshmp.netcdf.data.NetcdfShape.IndexKey;
import gov.usgs.earthquake.nshmp.netcdf.data.StaticData;

import ucar.ma2.Array;

/**
 * Abstract class to read in NetCDF file and create the bounding locations and
 * associated data.
 *
 * @author U.S. Geological Survey
 */
public abstract class BoundingReader<T> {

  private final NetcdfData netcdfData;
  private BoundingData<T> boundingData;
  private List<BoundingLocation> boundingLocations = new ArrayList<>();

  BoundingReader(Netcdf<T> netcdf, Location site) {
    this.netcdfData = netcdf.netcdfData();
    boundingLocations = setBoundingLocations(netcdf, site);
    boundingData = setBoundingData(netcdf, site, boundingLocations);
  }

  /**
   * Returns the bounding data associated with a location
   */
  public BoundingData<T> boundingData() {
    return boundingData;
  }

  /**
   * Returns the bounding locations
   */
  LocationList boundingLocations() {
    var locations = boundingLocations.stream()
        .map(boundingLocation -> boundingLocation.location)
        .collect(Collectors.toList());

    return LocationList.copyOf(locations);
  }

  abstract StaticData<T> calculateTargetData(
      List<BoundingLocation> boundingLocations,
      BoundingData<T> boundingData,
      double fracLon,
      double fracLat);

  abstract BoundingData<T> extractDataAt(
      Netcdf<T> netcdf,
      List<BoundingLocation> boundingLocations,
      int idxLonLL,
      int idxLatLL);

  /**
   * Get data for target point
   *
   * @param d1 data at first point (p1)
   * @param d2 data at second point (p2)
   * @param frac fractional distance between p1 and p2 to target point
   */
  abstract StaticData<T> getTargetData(StaticData<T> d1, StaticData<T> d2, double frac);

  /*
   * Read hazard curves from netCDF variable into map of hazards by SiteClass
   * and Imt
   *
   * TODO: if target is on a grid point (or on a grid lat or lon), no need to
   * read 4 bounding points ?
   */
  abstract StaticData<T> mapDataFromArray(
      Netcdf<T> netcdf,
      Array array);

  /**
   * Set the bounding data
   */
  abstract BoundingData<T> setBoundingData(
      Netcdf<T> netcdf,
      Location site,
      List<BoundingLocation> boundingLocations);

  static class BoundingLocation {
    final Location location;
    final int[] origin;

    BoundingLocation(
        NetcdfShape netcdfShape,
        Location location,
        int longitudeIndex,
        int latitudeIndex) {
      this.location = location;
      origin = netcdfShape.buildShape()
          .add(IndexKey.LATITUDE, latitudeIndex)
          .add(IndexKey.LONGITUDE, longitudeIndex)
          .build();
    }
  }

  private List<BoundingLocation> setBoundingLocations(Netcdf<T> netcdf, Location site) {
    var boundingLocations = new ArrayList<BoundingLocation>();
    var longitudes = netcdfData.longitudes();
    var latitudes = netcdfData.latitudes();

    var idxLonLL = NetcdfUtils.getIdxLTEQ(longitudes, site.longitude);
    var idxLatLL = NetcdfUtils.getIdxLTEQ(latitudes, site.latitude);

    var lonLeft = longitudes[idxLonLL];
    var lonRight = longitudes[idxLonLL + 1];
    var latLower = latitudes[idxLatLL];
    var latUpper = latitudes[idxLatLL + 1];

    boundingLocations.add(
        new BoundingLocation(netcdf.netcdfShape(), Location.create(lonLeft, latLower), 0, 0));
    boundingLocations.add(
        new BoundingLocation(netcdf.netcdfShape(), Location.create(lonLeft, latUpper), 0, 1));
    boundingLocations.add(
        new BoundingLocation(netcdf.netcdfShape(), Location.create(lonRight, latUpper), 1, 1));
    boundingLocations.add(
        new BoundingLocation(netcdf.netcdfShape(), Location.create(lonRight, latLower), 1, 0));

    return boundingLocations;
  }
}
