package gov.usgs.earthquake.nshmp.netcdf.www;

import static gov.usgs.earthquake.nshmp.netcdf.www.NetcdfWsUtils.GSON;

import java.util.Map;
import java.util.logging.Logger;

import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.gmm.Imt;
import gov.usgs.earthquake.nshmp.gmm.NehrpSiteClass;
import gov.usgs.earthquake.nshmp.netcdf.Netcdf;
import gov.usgs.earthquake.nshmp.netcdf.data.ScienceBaseMetadata;
import gov.usgs.earthquake.nshmp.netcdf.www.Query.Service;
import gov.usgs.earthquake.nshmp.netcdf.www.meta.DoubleParameter;
import gov.usgs.earthquake.nshmp.www.ResponseBody;

import io.micronaut.http.HttpRequest;
import io.micronaut.http.HttpResponse;
import io.swagger.v3.oas.annotations.media.Schema;

/**
 * Abstract service handler for {@code NetcdfController}.
 *
 * @see NetcdfController
 *
 * @author U.S. Geological Survey
 */
public abstract class NetcdfService<T extends Query> {

  protected static final Logger LOGGER = Logger.getLogger(NetcdfService.class.getName());

  Netcdf<?> netcdf;

  protected NetcdfService(Netcdf<?> netcdf) {
    this.netcdf = netcdf;
  }

  /**
   * Returns the metadata response.
   *
   * @param httpRequest The HTTP request
   */
  abstract ResponseBody<String, Metadata> getMetadataResponse(HttpRequest<?> httpRequest);

  /**
   * Returns the service name
   */
  abstract String getServiceName();

  /**
   * Returns the source model
   */
  abstract SourceModel getSourceModel();

  /**
   * Returns the netcdf object associated with the specific data type.
   */
  abstract Netcdf<?> netcdf();

  /**
   * Returns the static curves at a specific location.
   *
   * @param <T> The response type
   * @param request The request data
   * @param url The URL for the service call
   */
  abstract <U> ResponseBody<RequestData, U> processCurves(RequestData request, String url);

  /**
   * Returns the static curves at a specific location and site class.
   *
   * @param <T> The response type
   * @param request The request data
   * @param url The URL for the service call
   */
  abstract <U> ResponseBody<RequestDataSiteClass, U> processCurvesSiteClass(
      RequestDataSiteClass request,
      String url);

  /**
   * Process the service request and returns the reponse.
   *
   * @param httpRequest The HTTP request
   * @param query The HTTP query
   * @param service The NetCDF service
   */
  abstract ResponseBody<?, ?> processRequest(
      HttpRequest<?> httpRequest,
      T query,
      Service service);

  /**
   * Handler of HTTP request, returns the service response.
   *
   * @param httpRequest The HTTP request
   * @param query The service query
   */
  public HttpResponse<String> handleServiceCall(
      HttpRequest<?> httpRequest,
      T query) {
    try {
      var url = httpRequest.getUri().toString();
      LOGGER.info("Request: " + url);
      LOGGER.fine("Query:\n" + GSON.toJson(query));

      if (query.longitude == null && query.latitude == null) {
        return metadata(httpRequest);
      }
      var service = getService(query);
      var svcResponse = processRequest(httpRequest, query, service);
      var response = GSON.toJson(svcResponse);
      LOGGER.fine("Result:\n" + response);
      return HttpResponse.ok(response);
    } catch (Exception e) {
      var url = httpRequest.getUri().toString();
      return NetcdfWsUtils.handleError(e, getServiceName(), url);
    }
  }

  public Service getService(T query) {
    if (query.siteClass != null) {
      return Service.CURVES_BY_SITE_CLASS;
    } else {
      return Service.CURVES;
    }
  }

  private HttpResponse<String> metadata(HttpRequest<?> httpRequest) {
    var svcResponse = getMetadataResponse(httpRequest);
    var response = GSON.toJson(svcResponse);
    LOGGER.fine("Result:\n" + response);
    return HttpResponse.ok(response);
  }

  class Metadata {
    public final String description;
    public final String[] syntax;
    public final SourceModel model;
    public final DoubleParameter longitude;
    public final DoubleParameter latitude;
    public final DoubleParameter vs30;
    public final NetcdfMetadata netcdfMetadata;

    Metadata(HttpRequest<?> request, String description) {
      var url = request.getUri().toString();
      url = url.endsWith("/") ? url.substring(0, url.length() - 1) : url;
      this.description = description;
      syntax = new String[] {
          url + "/{longitude:number}/{latitude:number}",
          url + "?longitude={number}&latitude={number}",
          url + "/{longitude:number}/{latitude:number}/{siteClass:NehrpSiteClass}",
          url + "?longitude={number}&latitude={number}&siteClass={NehrpSiteClass}",
      };

      var min = netcdf().netcdfData().minimumBounds();
      var max = netcdf.netcdfData().maximumBounds();

      longitude = new DoubleParameter(
          "Longitude",
          "°",
          min.longitude,
          max.longitude);
      latitude = new DoubleParameter(
          "Latitude",
          "°",
          min.latitude,
          max.latitude);
      model = getSourceModel();
      vs30 = new DoubleParameter(
          "Vs30",
          "m/s",
          150,
          1500);
      netcdfMetadata = new NetcdfMetadata();
    }
  }

  class NetcdfMetadata {
    public final String netcdfFile;
    public final ScienceBaseMetadata scienceBaseMetadata;

    NetcdfMetadata() {
      var fileName = netcdf.netcdfPath().getFileName();
      netcdfFile = fileName == null ? netcdf().netcdfPath().toString() : fileName.toString();
      scienceBaseMetadata = netcdf().netcdfData().scienceBaseMetadata();
    }
  }

  class SourceModel {
    public final String name;
    public final Map<NehrpSiteClass, Double> siteClasses;

    SourceModel() {
      name = netcdf().netcdfData().scienceBaseMetadata().label;
      siteClasses = netcdf().netcdfData().vs30Map();
    }
  }

  static class ServiceResponseMetadata {
    public NehrpSiteClass siteClass;
    public String xLabel;
    public String yLabel;

    ServiceResponseMetadata(
        NehrpSiteClass siteClass,
        String xLabel,
        String yLabel) {
      this.siteClass = siteClass;
      this.xLabel = xLabel;
      this.yLabel = yLabel;
    }
  }

  static class ResponseData<T extends ServiceResponseMetadata> {
    final T metadata;
    final XySequence data;

    ResponseData(T metadata, XySequence data) {
      this.metadata = metadata;
      this.data = data;
    }

    public T getMetadata() {
      return metadata;
    }

    @Schema(implementation = XySequenceSchema.class)
    public XySequence getData() {
      return data;
    }
  }

  static class RequestData {
    public Double latitude;
    public Double longitude;
    public transient Location site;

    RequestData(Location site) {
      latitude = site.latitude;
      longitude = site.longitude;
      this.site = site;
    }
  }

  static class RequestDataImt extends RequestDataSiteClass {
    public Imt imt;

    RequestDataImt(Location site, NehrpSiteClass siteClass, Imt imt) {
      super(site, siteClass);
      this.imt = imt;
    }
  }

  static class RequestDataSiteClass extends RequestData {
    public NehrpSiteClass siteClass;

    RequestDataSiteClass(Location site, NehrpSiteClass siteClass) {
      super(site);
      this.siteClass = siteClass;
    }
  }

  private static class XySequenceSchema {
    public double[] xs;
    public double[] ys;
  }
}
